"""Tests the Map object for provision of geospatial data and map manipulation operations."""
import csv
import logging
import os
import shutil
import subprocess
import sys
import unittest

try:
    from osgeo import gdal
except ImportError as ie:
    try:
        import gdal
    except ImportError:
        raise ie
import numpy as np
from osgeo import osr, ogr, gdal

try:
    from unittest.mock import MagicMock, create_autospec, patch
except ImportError:  # pragma: no cover
    from mock import Mock as MagicMock
    from mock import create_autospec, patch

from pycoalescence import Map, FragmentedLandscape
from pycoalescence.map import shapefile_from_wkt, GdalErrorHandler
from setup_tests import setUpAll, tearDownAll, skipGdalWarp

import scipy

try:  # pragma: no cover
    from importlib import reload
except ImportError:  # pragma: no cover
    pass


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


@unittest.skipIf(sys.version[0] == "2", "Skipping Python 3.x tests")
class TestMapImports(unittest.TestCase):
    """Tests that errors are thrown correctly when gdal is not detected."""

    @classmethod
    def setUpClass(cls):
        """Mocks a fake subprocess call"""
        cls.orig_call = subprocess.check_output
        if "GDAL_DATA" in os.environ:
            cls.gdal_dir = os.environ["GDAL_DATA"]
            os.environ.pop("GDAL_DATA")
        else:  # pragma: no cover
            cls.gdal_dir = None

    @classmethod
    def tearDownClass(cls):
        """Re imports the subprocess module to remove the MagicMocking."""
        subprocess.check_output = cls.orig_call
        if cls.gdal_dir:
            os.environ["GDAL_DATA"] = cls.gdal_dir

    def testRaisesImportError(self):
        """Mocks a fake subprocess call to ``gdal-config --datadir`` and checks that the correct errors are raised"""
        subprocess.check_output = create_autospec(subprocess.check_output, return_value=b"/not/a/gdal/dir\n")
        import pycoalescence.map

        with self.assertRaises(ImportError):
            reload(pycoalescence.map)
        subprocess.check_output = self.orig_call
        subprocess.check_output = create_autospec(subprocess.check_output, return_value=None)
        with self.assertRaises(ImportError):
            reload(pycoalescence.map)


class TestMap(unittest.TestCase):
    """
    Tests the functions for the Map object work properly.
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the Map object
        """
        cls.fine_map = Map()
        cls.fine_map.file_name = "sample/SA_sample_fine.tif"
        cls.coarse_map = Map()
        cls.coarse_map.file_name = "sample/SA_sample_coarse.tif"
        cls.fine_offset_map = Map("sample/SA_sample_fine_offset.tif")
        cls.expected_WGS_84 = [
            'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]',
            'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],AUTHORITY["EPSG","4326"]]',
            'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST],AUTHORITY["EPSG","4326"]]',
        ]

    def testDetectGeoTransform(self):
        """Tests the fine and coarse map geo transforms."""
        # Test the fine map
        x, y, x_offset, y_offset, xres, yres, ulx, uly = self.fine_map.get_dimensions()
        self.assertEqual(x, 13)
        self.assertEqual(y, 13)
        self.assertAlmostEqual(ulx, -78.375, 8)
        self.assertAlmostEqual(uly, 0.8583333333333, 8)
        self.assertAlmostEqual(xres, 0.00833333333333, 8)
        self.assertAlmostEqual(yres, -0.0083333333333333, 8)
        self.assertEqual(0, x_offset)
        self.assertEqual(0, y_offset)
        x, y, x_offset, y_offset, xres, yres, ulx, uly = self.fine_map.get_dimensions()
        self.assertEqual(x, 13)
        self.assertEqual(y, 13)
        self.assertAlmostEqual(ulx, -78.375, 8)
        self.assertAlmostEqual(uly, 0.8583333333333, 8)
        self.assertAlmostEqual(xres, 0.00833333333333, 8)
        self.assertAlmostEqual(yres, -0.0083333333333333, 8)
        self.assertEqual(0, x_offset)
        self.assertEqual(0, y_offset)
        # Now test the coarse map
        x, y, x_offset, y_offset, xres, yres, ulx, uly = self.coarse_map.get_dimensions()
        self.assertEqual(x, 35)
        self.assertEqual(y, 41)
        self.assertAlmostEqual(ulx, -78.466666666, 8)
        self.assertAlmostEqual(uly, 0.975, 8)
        self.assertAlmostEqual(xres, 0.00833333333333, 8)
        self.assertAlmostEqual(yres, -0.0083333333333333, 8)
        self.assertEqual(0, x_offset)
        self.assertEqual(0, y_offset)

    def testCountBands(self):
        """Tests that the number of bands is correctly identified"""
        self.assertEqual(1, self.fine_map.get_band_number())
        multi_band = Map(os.path.join("sample", "multi_band.tif"))
        self.assertEqual(201, multi_band.get_band_number())

    def testGetXY(self):
        """Tests the raster dimensions are correctly identified."""
        x, y = self.fine_map.get_x_y()
        self.assertEqual(x, 13)
        self.assertEqual(y, 13)
        x, y = self.coarse_map.get_x_y()
        self.assertEqual(x, 35)
        self.assertEqual(y, 41)

    def testSetDimensions(self):
        """
        Tests set_dimensions method
        """
        tmp = Map()
        tmp.set_dimensions("sample/SA_sample_fine.tif")
        x, y, ulx, uly, xres, yres = [
            tmp.x_size,
            tmp.y_size,
            tmp.x_ul,
            tmp.y_ul,
            tmp.x_res,
            tmp.y_res,
        ]
        self.assertEqual(x, 13)
        self.assertEqual(y, 13)
        self.assertAlmostEqual(ulx, -78.375, 8)
        self.assertAlmostEqual(uly, 0.8583333333333, 8)
        self.assertAlmostEqual(xres, 0.00833333333333, 8)
        self.assertAlmostEqual(yres, -0.0083333333333333, 8)

    def testFail(self):
        """
        Tests the correct exceptions are thrown when stupid things are attempted!
        """
        map_fail = Map()
        map_fail.file_name = "sample/not_here.tif"

        with self.assertRaises(IOError):
            map_fail.get_dimensions()
        map_fail.file_name = "sample/SA_sample_fine"
        with self.assertRaises(IOError):
            map_fail.get_dimensions()
        map_fail.file_name = None
        with self.assertRaises(RuntimeError):
            map_fail.set_dimensions()
        with self.assertRaises(RuntimeError):
            map_fail.check_map()

    def testHasEqualDims(self):
        """
        Tests that dimensions are correctly identified when they do or do not match.
        """
        self.assertFalse(self.fine_map.has_equal_dimensions(self.coarse_map))
        self.assertFalse(self.fine_offset_map.has_equal_dimensions(self.fine_map))
        self.assertFalse(self.coarse_map.has_equal_dimensions(self.fine_map))
        m = Map("sample/SA_samplemaskINT.tif")
        self.assertTrue(self.fine_map.has_equal_dimensions(m))
        with self.assertRaises(TypeError):
            self.fine_map.has_equal_dimensions(10)
        m.x_size = "str"
        with self.assertRaises(ValueError):
            m.check_map()

    def testOffset(self):
        """Tests that the offsets are correctly calculated between the fine and coarse maps."""
        self.assertListEqual(self.coarse_map.calculate_offset(self.fine_map), [-11, -14])
        self.assertListEqual(self.coarse_map.calculate_offset(self.fine_map.file_name), [-11, -14])
        self.assertListEqual(self.coarse_map.calculate_offset(self.coarse_map), [0, 0])
        self.assertListEqual(self.fine_map.calculate_offset(self.coarse_map), [11, 14])

    def testScale(self):
        """Tests that the scale is correctly calculated."""
        self.assertEqual(1, self.fine_map.calculate_scale(self.fine_map))
        self.assertEqual(1, self.fine_map.calculate_scale(self.fine_map.file_name))
        self.assertEqual(
            12.99959999999942,
            self.coarse_map.calculate_scale(os.path.join("sample", "null_coarse.tif")),
        )
        with self.assertRaises(ValueError):
            _ = self.coarse_map.calculate_scale(10)

    def testSample(self):
        m = Map(os.path.join("sample", "SA_sample_fine.tif"))
        m.set_sample(True)
        self.assertTrue(m.is_sample)
        self.assertEqual(0, m.x_offset)
        self.assertEqual(0, m.y_offset)

    def testExtents(self):
        """Tests that the extents are correctly calculated for the Map files."""
        self.assertListEqual(
            [
                -78.3750000000000000,
                -78.2666666666666657,
                0.7500000000000009,
                0.8583333333333343,
            ],
            self.fine_map.get_extent(),
        )

    def testIsWithin(self):
        """
        Tests that the maps are correctly identified as within or outside of each other.
        """
        self.assertTrue(self.fine_map.is_within(self.coarse_map))
        self.assertFalse(self.fine_map.is_within(self.fine_offset_map))
        self.assertFalse(self.fine_offset_map.is_within(self.fine_map))
        self.assertFalse(self.coarse_map.is_within(self.fine_map))
        with self.assertRaises(TypeError):
            self.fine_map.is_within(0.111)

    def testCoordinateConversion(self):
        """
        Tests that coordinates are correctly converted between cooridinate types
        """
        m = Map("sample/SA_sample_fine.tif")
        x, y = m.convert_lat_long(0.858333, -78.375)
        self.assertEqual(x, 0)
        self.assertEqual(y, 0)
        x, y = m.convert_lat_long(0.854113550519, -78.3114884325)
        self.assertEqual(x, 7)
        self.assertEqual(y, 0)
        x, y = m.convert_lat_long(0.770653918791, -78.3032713645)
        self.assertEqual(x, 8)
        self.assertEqual(y, 10)

    def testGetSpatialReference(self):
        m = Map("sample/SA_sample_fine.tif")
        proj = m.get_projection()
        self.assertIn(proj, self.expected_WGS_84)

    def testRasteriseShapefile(self):
        """
        Tests basic rasterisation of a shapefile is successful.
        """
        m = Map()
        output_raster = "output/raster_out1.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([13, 71, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(361, np.sum(m.data))
        self.assertEqual(1, m.data[20, 4])
        self.assertEqual(0, m.data[20, 6])

    def testRasteriseShapefileWithExtent(self):
        """
        Tests rasterisation using an extent is successful.
        """
        m = Map()
        output_raster = "output/raster_out1b.tif"
        new_extent = [-78.363529863, -78.321863213, 0.817435788, 0.850435788]

        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            extent=new_extent,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([4, 32, 0, 0, 0.00833333, -0.001, -78.36352986307837, 0.8504357884796987]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(73, np.sum(m.data))
        self.assertEqual(0, m.data[2, 2])
        self.assertEqual(1, m.data[10, 2])

    def testRasteriseShapefileWithFields(self):
        """
        Tests that values are correctly applied for rasterisation of a shapefile with fields.
        """
        m = Map()
        output_raster = "output/raster_out2.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            field="field1",
            burn_val=100,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([13, 71, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(4460, np.sum(m.data))
        self.assertEqual(10, m.data[21, 5])
        self.assertEqual(0, m.data[21, 7])
        self.assertEqual(20, m.data[13, 10])

    def testRasteriseShapefileWithFields2(self):
        """
        Tests that values are correctly applied for rasterisation of a shapefile with fields.
        """
        m = Map()
        output_raster = "output/raster_out3.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            field="field2",
            burn_val=100,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([13, 71, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(67705, np.sum(m.data))
        self.assertEqual(245, m.data[20, 4])
        self.assertEqual(0, m.data[21, 7])
        self.assertEqual(1, m.data[13, 10])

    def testRasteriseShapefileWithFields3(self):
        """
        Tests that values are correctly applied for rasterisation of a shapefile with fields.
        """
        m = Map()
        output_raster = "output/raster_out4.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            burn_val=100,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([13, 71, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(36100, np.sum(m.data))
        self.assertEqual(100, m.data[20, 4])
        self.assertEqual(0, m.data[21, 7])
        self.assertEqual(100, m.data[13, 10])

    def testRasteriseShapefileWithFields4(self):
        """
        Tests that values are correctly applied for rasterisation of a shapefile with selection of one particular field
        """
        m = Map()
        output_raster = "output/raster_out5.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            attribute_filter="field1=10",
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([10, 65, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(274, np.sum(m.data))
        self.assertEqual(1, m.data[1, 4])

    def testRasteriseShapefileWithBuffer(self):
        """
        Tests that the shapefile can be rasterised with an extra buffer around the edge.
        """
        m = Map()
        output_raster = "output/raster_out6.tif"
        m.rasterise(
            shape_file="sample/shape_sample.shp",
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            x_buffer=2.0,
            y_buffer=3.0,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([15, 75, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        self.assertEqual(361, np.sum(m.data))
        self.assertEqual(1, m.data[20, 4])
        self.assertEqual(0, m.data[21, 7])

    def testRasteriseShapefileWithDataSource(self):
        """
        Tests that a data source can be used for rasterising the shapefile instead of a file path.
        """
        m = Map()
        output_raster = "output/raster_out7.tif"
        output_ds = ogr.Open("sample/shape_sample.shp")
        m.rasterise(
            shape_file=output_ds,
            raster_file=output_raster,
            x_res=0.00833333,
            y_res=0.001,
            x_buffer=2.0,
            y_buffer=3.0,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([15, 75, 0, 0, 0.00833333, -0.001]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        output_ds = None
        self.assertEqual(361, np.sum(m.data))
        self.assertEqual(1, m.data[20, 4])
        self.assertEqual(0, m.data[21, 7])

    def testRasteriseShapefileWithGeotransform(self):
        """
        Tests that a custom geotransform can be provided for during rasterisation.
        """
        m = Map()
        output_raster = "output/raster_out8.tif"
        geo_transform = (-78.375, 0.00833333, 0, 0.858333, 0, -0.00833333)
        output_ds = ogr.Open("sample/shape_sample.shp")
        m.rasterise(shape_file=output_ds, raster_file=output_raster, geo_transform=geo_transform)
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([12, 10, 0, 0, 0.00833333, -0.00833333]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        output_ds = None
        self.assertEqual(53, np.sum(m.data))
        self.assertEqual(1, m.data[5, 4])
        self.assertEqual(0, m.data[0, 0])

    def testRasteriseShapefileWithGeotransformAndDims(self):
        """
        Tests that a custom geotransform and dimensions can be provided for during rasterisation.
        """
        m = Map()
        output_raster = "output/raster_out9.tif"
        geo_transform = (-78.375, 0.00833333, 0, 0.858333, 0, -0.00833333)
        output_ds = ogr.Open("sample/shape_sample.shp")
        m.rasterise(
            shape_file=output_ds,
            raster_file=output_raster,
            geo_transform=geo_transform,
            width=13,
            height=13,
        )
        self.assertTrue(os.path.exists(output_raster))
        dims = m.get_dimensions()
        for i, each in enumerate([13, 13, 0, 0, 0.00833333, -0.00833333]):
            self.assertAlmostEqual(each, dims[i], places=5)
        self.assertIn(m.get_projection(), self.expected_WGS_84)
        m.open()
        output_ds = None
        self.assertEqual(53, np.sum(m.data))
        self.assertEqual(1, m.data[5, 4])
        self.assertEqual(0, m.data[0, 0])

    def testRasteriseErrors(self):
        """
        Tests that rasterise raises the correct errors when files don't exist, or incorrect parameters are provided.
        """
        m = Map()
        with self.assertRaises(IOError):
            m.rasterise(
                shape_file="not_exist",
                raster_file="output.tif",
                x_res=0.00833333,
                y_res=0.001,
                field="field2",
                burn_val=100,
            )
        m = Map()
        with self.assertRaises(IOError):
            m.rasterise(
                shape_file="sample/shape_sample.shp",
                raster_file="sample/SA_sample.tif",
                x_res=0.00833333,
                y_res=0.001,
                field="field2",
                burn_val=100,
            )
        with self.assertRaises(ValueError):
            m.rasterise(
                shape_file="sample/SA_sample.tif",
                raster_file="output.tif",
                x_res=0.00833333,
                y_res=0.001,
                field="field2",
                burn_val=100,
            )
        m = Map()
        with self.assertRaises(ValueError):
            m.rasterise(shape_file="sample/shape_sample.shp", raster_file="no_exist.tif")
        m = Map()
        with self.assertRaises(ValueError):
            m.rasterise(
                shape_file="sample/shape_sample.shp",
                raster_file="output.tif",
                geo_transform=(1, 1, 1, 1, 1, 1),
                x_res=1.0,
                y_res=1.0,
            )

    def testCreateCopy(self):
        """
        Tests that create copy works as intended.
        """
        m = Map()
        src_file = os.path.join("sample", "SA_sample.tif")
        dst_file = os.path.join("output", "copy.tif")
        m2 = Map(src_file)
        m.create_copy(dst_file=dst_file, src_file=src_file)
        self.assertTrue(os.path.exists(dst_file))
        m.file_name = src_file
        with self.assertRaises(IOError):
            m.create_copy(dst_file=dst_file)
        self.assertListEqual(m.get_dimensions(), m2.get_dimensions())
        self.assertEqual(m.get_projection(), m2.get_projection())

    def testCreate(self):
        """
        Tests that create function works properly.
        """
        m = Map()
        output_file = "output/test_create.tif"
        m.data = np.zeros(shape=(10, 10))
        m.create(output_file)
        self.assertTrue(os.path.exists(output_file))
        m2 = Map(output_file)
        m2.open()
        self.assertEqual((10, 10), m2.data.shape)
        self.assertEqual(0, np.sum(m2.data))
        with self.assertRaises(IOError):
            m.create(output_file)
        m.data = None
        with self.assertRaises(ValueError):
            m.create(output_file)

    def testCreateWithGeotransform(self):
        """
        Tests that the create function works properly with setting a geotransform.
        """
        m = Map()
        np.random.seed(1)
        m.data = np.random.rand(10, 10)
        output_file = "output/test_create2.tif"
        gt = (
            -76.54166666666666,
            0.008333333333333333,
            0.0,
            5.024999999999999,
            0.0,
            -0.008333333333333333,
        )
        proj = (
            'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],'
            'AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],AUTHORITY["EPSG",'
            '"4326"]]'
        )
        m.create(
            output_file,
            bands=1,
            datatype=gdal.GDT_Float32,
            geotransform=gt,
            projection=proj,
        )
        m2 = Map(output_file)
        m2.open()
        self.assertEqual((10, 10), m2.data.shape)
        self.assertAlmostEqual(48.58779, np.sum(m2.data), places=5)

    def testDataTypeFetch(self):
        """Tests that the data type is correctly obtained."""
        m = Map("sample/SA_sample.tif")
        self.assertEqual(6, m.get_dtype())
        self.assertEqual(6, m.get_dtype(band_no=1))

    def testGetDataset(self):
        """Tests that the dataset is obtained correctly."""
        m = Map(os.path.join("sample", "PlotBiodiversityMetrics.db"))
        with self.assertRaises(IOError):
            m.get_dataset()

    def testCalculateOffset(self):
        """Tests that the calculate offset errors work as intended."""
        m = Map(os.path.join("sample", "SA_sample_fine.tif"))
        m2 = Map(os.path.join("sample", "SA_sample_coarse.tif"))
        x, y = m.calculate_offset(m2)
        self.assertEqual(11, x)
        self.assertEqual(14, y)
        with self.assertRaises(TypeError):
            m.calculate_offset(10)

    def get_null_projection(self):
        """Get null projection for mock testing."""
        a = 0
        while True:
            a += 1
            yield a

    @patch.object(Map, "get_projection", get_null_projection)
    def testCalculateOffsetMixedProjection(self):
        """Tests that the calculate offset errors work as intended."""
        m = Map(os.path.join("sample", "SA_sample_fine.tif"), logging_level=50)
        m2 = Map(os.path.join("sample", "SA_sample_coarse.tif"), logging_level=50)
        with self.assertRaises(TypeError):
            _, _ = m.calculate_offset(m2)

    def testTranslate(self):
        """Tests the translation process works as intended."""
        source_file = os.path.join("sample", "SA_sample_fine.tif")
        dest_file1 = os.path.join("output", "translated1.tif")
        dest_file2 = os.path.join("output", "translated2.tif")
        m = Map(source_file)
        new_x, new_y = 10, 10
        new_proj_win = [-78.375, 0.8583333333333343, -78.3, 0.8]
        expected_extent = [-78.375, -78.3, 0.8, 0.8583333333333343]
        m.translate(dest_file=dest_file1, projWin=new_proj_win)
        with self.assertRaises(ValueError):
            m.translate(dest_file=None, projWin=new_proj_win)
        with self.assertRaises(IOError):
            m.translate(dest_file=dest_file1, projWin=new_proj_win)
        m = Map()
        m.translate(source_file=source_file, dest_file=dest_file2, width=new_x, height=new_y)
        self.assertTrue(os.path.exists(dest_file1))
        m1 = Map(dest_file1)
        actual_extent = m1.get_extent()
        for expected, actual in zip(expected_extent, actual_extent):
            self.assertAlmostEqual(expected, actual, places=5)
        self.assertTrue(os.path.exists(dest_file1))
        m2 = Map(dest_file2)
        self.assertEqual([new_x, new_y], m2.get_x_y())


@unittest.skipUnless(hasattr(gdal, "Warp"), "Skipping reprojection test as gdal.Warp not found.")
@skipGdalWarp
class TestMapReprojection(unittest.TestCase):
    """
    Tests that reprojections work as intended.
    """

    @classmethod
    def setUpClass(cls):
        """
        Performs the initial re-projections.
        """
        cls.map_0 = os.path.join("sample", "multi_band.tif")
        cls.map_1 = os.path.join("sample", "SA_sample.tif")
        cls.map_2 = os.path.join("output", "tmp_map1.tif")
        cls.map_3 = os.path.join("output", "tmp_map2.tif")
        cls.map_4 = os.path.join("output", "tmp_map3.tif")
        cls.map_5 = os.path.join("output", "tmp_map5.tif")
        cls.m0 = Map(cls.map_0)
        cls.m = Map(cls.map_1)
        cls.destination_projection = osr.SpatialReference()
        res = cls.destination_projection.ImportFromEPSG(3857)
        if res != 0:
            raise RuntimeError("Could not import from EPSG in osr.SpatialReference: {}.".format(res))
        cls.proj_wkt = cls.destination_projection.ExportToWkt()
        cls.m.reproject_raster(dest_file=cls.map_2, dest_projection=cls.destination_projection)
        cls.m.reproject_raster(
            dest_file=cls.map_3,
            dest_projection=cls.destination_projection,
            x_scalar=2,
            y_scalar=10,
        )
        cls.m0.reproject_raster(dest_file=cls.map_5, x_scalar=0.1, y_scalar=0.1)
        cls.m_2 = Map(cls.map_2)
        cls.m_3 = Map(cls.map_3)
        cls.m_5 = Map(cls.map_5)

    def testCorrectNewProjection(self):
        """
        Tests that reprojection works correctly, either creating a new file or over-writing the original file
        """

        self.assertNotEqual(self.m.get_projection(), self.proj_wkt)
        self.assertEqual(self.m0.get_projection(), self.m_5.get_projection())
        self.assertEqual(self.proj_wkt, self.m_2.get_projection())
        self.assertEqual(self.proj_wkt, self.m_3.get_projection())
        self.assertEqual(self.proj_wkt, self.m_3.get_projection())

    def testNewDimensions(self):
        """
        Tests that the new dimensions are correctly set
        """

        dims_0 = self.m0.get_dimensions()
        dims_1 = self.m.get_dimensions()
        dims_2 = self.m_2.get_dimensions()
        dims_3 = self.m_3.get_dimensions()
        dims_5 = self.m_5.get_dimensions()
        self.assertListEqual(dims_1[:4], dims_2[:4])
        self.assertAlmostEqual(929.3830958815255, dims_2[4], places=5)
        self.assertAlmostEqual(-929.3830958815255, dims_2[5], places=5)
        self.assertAlmostEqual(-8520579.357801815, dims_2[6], places=5)
        self.assertAlmostEqual(560098.928515636, dims_2[7], places=5)
        self.assertAlmostEqual(dims_2[4] * 2, dims_3[4], places=5)
        self.assertAlmostEqual(dims_2[5] * 10, dims_3[5], places=5)
        self.assertAlmostEqual(dims_0[0] * 10, dims_5[0], places=5)
        self.assertAlmostEqual(dims_0[1] * 10, dims_5[1], places=5)
        self.assertAlmostEqual(dims_0[4] * 0.1, dims_5[4], places=5)
        self.assertAlmostEqual(dims_0[5] * 0.1, dims_5[5], places=5)
        shutil.copy2(self.map_1, self.map_4)
        m_4 = Map(self.map_4)
        m_4.reproject_raster(dest_projection=self.destination_projection)
        self.assertListEqual(m_4.read_dimensions(), dims_2)
        self.assertEqual(m_4.get_projection(), self.proj_wkt)

    def testNewValues(self):
        """
        Tests that the new values are correctly stored in the raster file
        """
        self.m.open()
        self.m_2.open()
        for i in range(self.m.data.shape[0]):
            for j in range(self.m.data.shape[1]):
                self.assertEqual(self.m.data[i, j], self.m_2.data[i, j])
        self.m_3.open()
        self.assertEqual(444, self.m_3.data[0, 0])
        self.assertEqual(258, self.m_3.data[1, 11])

    def testDownscaling(self):
        """
        Tests that re-scaling works without trying to reproject the raster.
        """
        map_6 = "output/tmp_map6.tif"
        shutil.copy2(self.map_2, map_6)
        m_6 = Map(map_6, logging_level=20)
        m_6.reproject_raster(x_scalar=2.0, y_scalar=3.0)
        m_4 = Map(self.map_2)
        dims = m_6.read_dimensions()
        dims_2 = m_4.get_dimensions()
        for i, each in enumerate(dims_2[6:]):
            self.assertAlmostEqual(each, dims[i + 6])
        self.assertAlmostEqual(dims[4], 2 * dims_2[4])
        self.assertAlmostEqual(dims[5], 3 * dims_2[5])

    def testUpScaling(self):
        """
        Tests that re-scaling works without trying to reproject the raster.
        """
        map_5 = "output/tmp_map5.tif"
        shutil.copy2(self.map_2, map_5)
        m_5 = Map(map_5)
        m_5.reproject_raster(x_scalar=0.5, y_scalar=0.5)
        m_4 = Map(self.map_2)
        dims = m_5.read_dimensions()
        dims_2 = m_4.get_dimensions()
        for i, each in enumerate(dims_2[6:]):
            self.assertAlmostEqual(each, dims[i + 6])
        self.assertAlmostEqual(dims[4], 0.5 * dims_2[4])
        self.assertAlmostEqual(dims[5], 0.5 * dims_2[5])

    def testNoData(self):
        """
        Tests that no data values are correctly read from tif files.
        """
        m = Map("sample/large_mask.tif")
        self.assertEqual(-99.0, m.get_no_data())
        self.assertEqual(-99.0, m.get_no_data(1))
        m2 = Map("sample/bytesample.tif")
        self.assertIsNone(m2.get_no_data())
        m3 = Map("sample/example_historical_fine.tif")
        self.assertAlmostEqual(-3.4028234663852886e38, m3.get_no_data(), places=7)

    def testMultiBandsExist(self):
        """Tests that the correct number of bands exist in the output raster."""
        self.assertEqual(201, self.m_5.get_band_number())
        # Test that the first 5 bands contain non-zero data (as this is what was in the original)
        for i in range(1, 5, 1):
            self.m_5.open(band_no=i)
            self.assertGreater(np.sum(self.m_5.data), 0)
            self.m0.open(band_no=i)
            self.assertAlmostEqual(np.sum(self.m0.data) * 100, np.sum(self.m_5.data), places=5)


class MapAssignment(unittest.TestCase):
    """
    Asserts that the Map class correctly reads and writes data properly.
    """

    @classmethod
    def setUpClass(cls):
        shutil.copy(os.path.join("sample", "null.tif"), "output")
        cls.map = Map(file=os.path.join("output", "null.tif"))
        cls.map.open()
        cls.map.data[0:5, 0:2] = 10
        cls.map.write(band_no=1)

    def testBaseMap(self):
        """
        Just a double check to make sure that the base map file is correct (not really a test of this code at all).
        """
        ds = gdal.Open(os.path.join("sample", "null.tif"))
        arr = ds.GetRasterBand(1).ReadAsArray()
        self.assertEqual(np.sum(arr), 169)
        ds = None

    def testMapUpdates(self):
        ds = gdal.Open(os.path.join("output", "null.tif"))
        arr = ds.GetRasterBand(1).ReadAsArray()
        self.assertEqual(np.sum(arr), 259)
        ds = None

    def testWriteError(self):
        """Tests that the write function raises the expected errors."""
        m = Map("none.tif")
        with self.assertRaises(IOError):
            m.write()
        m = Map(os.path.join("sample", "null.tif"))
        with self.assertRaises(ValueError):
            m.write()
        m = Map()
        with self.assertRaises(ValueError):
            m.map_exists()


@unittest.skipUnless(
    hasattr(scipy.spatial, "Voronoi"),
    "Skipping reprojection test as scipy.spatial.Voronoi not found.",
)
class TestFragmentedLandscape(unittest.TestCase):
    """
    Tests that the fragmented landscape generation creates successfully for a range of fragment numbers and sizes.
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the class object by creating the required maps
        :return:
        """
        if os.path.exists("landscapes"):
            shutil.rmtree("landscapes")
        cls.l1 = FragmentedLandscape(
            size=10,
            number_fragments=2,
            total=4,
            output_file=os.path.join("landscapes", "l1.tif"),
        )
        cls.l2 = FragmentedLandscape(
            size=100,
            number_fragments=57,
            total=150,
            output_file=os.path.join("landscapes", "l2.tif"),
        )
        cls.l3 = FragmentedLandscape(
            size=24,
            number_fragments=5,
            total=5,
            output_file=os.path.join("landscapes", "l3.tif"),
        )
        cls.l1.generate()
        cls.l2.generate()
        cls.l3.generate()
        cls.l4 = FragmentedLandscape(
            size=10,
            number_fragments=1,
            total=2,
            output_file=os.path.join("landscapes", "l4.tif"),
        )
        cls.l4.generate()
        cls.l5 = FragmentedLandscape(
            size=100,
            number_fragments=100,
            total=2000,
            output_file=os.path.join("landscapes", "l5.tif"),
        )
        cls.l5.generate()

    @classmethod
    def tearDownClass(cls):
        # pass
        if os.path.exists("landscapes"):
            shutil.rmtree("landscapes")

    def testCreateFragmentedLandscapes(self):
        self.assertEqual(self.l1.output_file, os.path.join("landscapes", "l1.tif"))
        self.assertTrue(os.path.exists(self.l1.output_file))
        self.assertEqual(self.l2.output_file, os.path.join("landscapes", "l2.tif"))
        self.assertTrue(os.path.exists(self.l2.output_file))
        self.assertEqual(self.l3.output_file, os.path.join("landscapes", "l3.tif"))
        self.assertTrue(os.path.exists(self.l3.output_file))
        self.assertEqual(self.l4.output_file, os.path.join("landscapes", "l4.tif"))
        self.assertTrue(os.path.exists(self.l4.output_file))

    def testDimensionsCorrect1(self):
        """
        Checks that the saved dimensions are correct.
        """
        m1 = Map(logging_level=logging.CRITICAL)
        m1.file_name = self.l1.output_file
        m1.set_dimensions()
        self.assertEqual(m1.x_size, 10)
        self.assertEqual(m1.y_size, 10)
        m1 = Map(logging_level=logging.CRITICAL)
        m1.file_name = self.l1.output_file
        m1.set_dimensions(x_size=10)
        self.assertEqual(m1.x_size, 10)
        self.assertEqual(m1.y_size, 10)

    def testDimensionsCorrect2(self):
        """
        Checks that the saved dimensions are correct.
        """
        m2 = Map()
        m2.file_name = self.l2.output_file
        m2.set_dimensions()
        self.assertEqual(m2.x_size, 100)
        self.assertEqual(m2.y_size, 100)

    def testDimensionsCorrect3(self):
        """
        Checks that the saved dimensions are correct.
        """
        m3 = Map()
        m3.file_name = self.l3.output_file
        m3.set_dimensions()
        self.assertEqual(m3.x_size, 24)
        self.assertEqual(m3.y_size, 24)

    def testDimensionsCorrect5(self):
        """
        Checks that the saved dimensions are correct.
        """
        m5 = Map()
        m5.file_name = self.l5.output_file
        m5.set_dimensions()
        self.assertEqual(m5.x_size, 100)
        self.assertEqual(m5.y_size, 100)


class TestSubsettingMaps(unittest.TestCase):
    """
    Tests the reading of tif files works correctly, including subsetting and cached-subsetting to obtain selections from
    the maps.
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the map object for reading from.
        """
        cls.m = Map("sample/SA_sample_fine.tif")

    def testSubsetting(self):
        """
        Tests the basic subsetting functionality works as intended.
        :return:
        """
        arr = self.m.get_subset(x_offset=0, y_offset=0, x_size=13, y_size=13)
        with self.assertRaises(ValueError):
            self.m.get_subset(-1, -1, 13, 13)
        with self.assertRaises(ValueError):
            self.m.get_subset(0, 0, 14, 14)
        with self.assertRaises(ValueError):
            self.m.get_subset(14, 14, 1, 1)
        self.assertEqual(arr[0, 0], 231)
        self.assertEqual(arr[1, 0], 296)
        self.assertEqual(arr[0, 1], 303)
        arr2 = self.m.get_subset(x_offset=5, y_offset=5, x_size=2, y_size=2)
        self.assertEqual(arr2[0, 0], 288)
        self.assertEqual(arr2[1, 0], 263)
        self.assertEqual(arr2[0, 1], 286)

    def testCachedSubsetting(self):
        """
        Tests that the cached subsetting functionality works as intended.
        :return:
        """
        arr = self.m.get_cached_subset(x_offset=0, y_offset=0, x_size=13, y_size=13)
        self.assertEqual(arr[0, 0], 231)
        self.assertEqual(arr[1, 0], 296)
        self.assertEqual(arr[0, 1], 303)
        arr2 = self.m.get_cached_subset(x_offset=5, y_offset=5, x_size=2, y_size=2)
        self.assertEqual(arr2[0, 0], 288)
        self.assertEqual(arr2[1, 0], 263)
        self.assertEqual(arr2[0, 1], 286)

    def testSubsettingNoData(self):
        """
        Tests that the correct array is returned when using a no data value.
        """
        m = Map("sample/large_mask.tif")
        self.assertEqual(-243335289.0, np.sum(m.get_subset(0, 0, 1818, 1695)))
        self.assertEqual(375, np.sum(m.get_subset(0, 0, 1818, 1695, 0)))

    def testWritingSubset(self):
        """Tests that writing a subset of the gdal raster works as intended."""
        m = Map()
        tmp_file = "output/writesubset.tif"
        m.data = np.ones((10, 10))
        m.create(tmp_file)
        m = Map(tmp_file)
        tmp_array = np.ones((5, 5)) + 1
        m.write_subset(tmp_array, 2, 2)
        m.file_name = "none.tif"
        with self.assertRaises(IOError):
            m.write_subset(tmp_array, 2, 2)
        tmp_array = np.ones((10000, 10000))
        m.file_name = tmp_file
        with self.assertRaises(ValueError):
            m.write_subset(tmp_array, 2, 2)

        m = Map(tmp_file)
        m.open()
        expected_array = np.ones((10, 10))
        expected_array[2:7, 2:7] = 2
        self.assertTrue(np.array_equal(expected_array, m.data))


class TestWKTFunctions(unittest.TestCase):
    """Tests converting WKTs in csvs to shapefiles"""

    def testDefaultShapefileFromWKT(self):
        """Tests the default conversion of WKT to shapefile."""
        output_shapefile1 = os.path.join("output", "wkt_shapefile1.shp")
        output_shapefile2 = os.path.join("output", "wkt_shapefile2.shp")
        with open(os.path.join("sample", "wkt_test.csv"), "r") as input_file:
            csv_reader = csv.DictReader(input_file)
            lines = [x for x in csv_reader]
            wkts = [[z for x, z in y.items() if x == "WKT"][0] for y in lines]
            shapefile_from_wkt(wkts=wkts, dest_file=output_shapefile1)
            shapefile_from_wkt(wkts=wkts, dest_file=output_shapefile2, fields=lines)
            with self.assertRaises(IOError):
                shapefile_from_wkt(wkts=wkts, dest_file=output_shapefile1)
        self.assertTrue(os.path.exists(output_shapefile1))
        self.assertTrue(os.path.exists(output_shapefile2))


class TestGdalErrorHandler(unittest.TestCase):
    def testErrorHandler(self):
        class TempLogger(object):
            """Test logger for testing outputting."""

            def __init__(self):
                self.logs = []

            def log(self, err_level, err_msg):
                self.logs.append([err_level, err_msg])

        tl = TempLogger()
        geh = GdalErrorHandler(tl)
        geh.handler(10, 1, "test message")
        geh.handler(20, 2, "test message 2")
        expected_output = [[10, "test message"], [20, "test message 2"]]
        self.assertEqual(expected_output, tl.logs)
        self.assertEqual("test message 2", geh.err_msg)
        self.assertEqual(20, geh.err_level)
        self.assertEqual(2, geh.err_no)



"""
Tests the mean nearest-neighbour dispersal calculator for map files.
"""

import unittest

from pycoalescence.landscape_metrics import LandscapeMetrics


class TestLandscapeMetrics(unittest.TestCase):
    """Tests the MNN calculator with null files, and for randomly distributed landscapes."""

    def testNullLandscape(self):
        """Tests that a null landscape generates a MNN of 1"""
        mnn = LandscapeMetrics("sample/null.tif", logging_level=50)
        self.assertEqual(1.0, mnn.get_mnn())

    def testEvenLandscape(self):
        """Tests that an even landscape generates a MNN of 2."""
        mnn = LandscapeMetrics("sample/even_landscape.tif", logging_level=50)
        self.assertEqual(2.0, mnn.get_mnn())

    def testSampleLandscape(self):
        """Tests that the sample landscape generates the expected mean nearest-neighbour distance."""
        mnn = LandscapeMetrics("sample/SA_sample.tif", logging_level=50)
        self.assertEqual(1.0, mnn.get_mnn())

    def testSampleLandscape2(self):
        """Tests that the sample landscape generates the expected mean nearest-neighbour distance."""
        mnn = LandscapeMetrics("sample/SA_samplemaskINT.tif", logging_level=50)
        self.assertAlmostEqual(1.00618, mnn.get_mnn(), places=5)

    def testClumpyDisaggregated(self):
        """Tests that clumpy produces the expected metric for a fully disaggregated landscape."""
        l = LandscapeMetrics("sample/even_landscape.tif", logging_level=50)
        self.assertEqual(-1.0, l.get_clumpiness())
        l = LandscapeMetrics("sample/null_sample.tif", logging_level=50)
        self.assertEqual(-1.0, l.get_clumpiness())

    def testClumpyAggregated(self):
        """Tests that clumpy produces the expected metric for an aggregated landscape."""
        l = LandscapeMetrics("sample/null.tif", logging_level=50)
        self.assertEqual(1.0, l.get_clumpiness())

    def testClumpyModerate(self):
        """Tests that clumpy produces the expected metric for a moderately aggregated landscape."""
        l = LandscapeMetrics("sample/SA_samplemaskINT.tif", logging_level=50)
        self.assertAlmostEqual(0.445512069369, l.get_clumpiness(), places=5)

    def testMultipleMetrics(self):
        """Tests that multiple landscape metrics can be obtained on the same object."""
        mnn = LandscapeMetrics("sample/null_sample.tif", logging_level=50)
        self.assertAlmostEqual(1.00618, mnn.get_mnn(), places=5)
        self.assertAlmostEqual(-1.0, mnn.get_clumpiness(), places=5)



"""
Compiles necsim for several simulation scenarios and tests the output to make sure simulation results are as expected.
Also performs limited tests of the pycoalescence setup routines.

.. note:: If test_install.py has not been run before, compilation may take a while, depending on your system, as it
   compiles NECSim for several different options.
"""
from __future__ import absolute_import

import argparse
import logging
import sys
import unittest

try:
    import sqlite3
except ImportError:
    import sqlite as sqlite3

try:
    import pycoalescence
    from pycoalescence.system_operations import set_logging_method
    import setup_tests
except ImportError as ie:
    logging.warning("Cannot import pycoalescence globally, check package is properly installed: {}.".format(ie))
    logging.warning("Continuing with local package")
    sys.path.append("../")

    # Conditional import for Python 2 being stupid
    from system_operations import set_logging_method
    import setup_tests
if sys.version_info[0] != 3:

    class FileExistsError(IOError):
        pass


def main(verbosity=1):
    """
    Set the logging method, run the program compilation (if required) and test the install.

    :param verbosity: the level of information to display from the unittest module

    .. note:: The working directory is changed to the package install location for the duration of this execution.
    """
    set_logging_method(logging_level=logging.CRITICAL, output=None)
    test_loader = unittest.TestLoader().discover(".")
    unittest.TextTestRunner(verbosity=verbosity).run(test_loader)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Test pycoalescence functions correctly")
    parser.add_argument("--quick", help="Run only quick tests.", action="store_true", default=False)
    parser.add_argument("--bypass_gdal_warp", help="Bypass gdal.Warp.", action="store_true", default=False)
    parser.add_argument("-v", "--verbose", help="Use verbose mode.", action="store_true", default=False)
    args, unknown = parser.parse_known_args()
    if args.quick:
        setup_tests.quick_test = True
        sys.argv.remove("--quick")
    else:
        setup_tests.quick_test = False
    if args.bypass_gdal_warp:
        setup_tests.bypass_gdal_warp = True
        sys.argv.remove("--bypass_gdal_warp")
    else:
        setup_tests.bypass_gdal_warp = False
    if args.verbose:
        main(2)
    else:
        main()



"""
Calculates landscape-level metrics, including mean distance to nearest-neighbour for each habitat cell and clumpiness.
"""
import logging

from pycoalescence.necsim.necsim import CLandscapeMetricsCalculator

try:
    from .map import Map
    from .system_operations import write_to_log
except ImportError:  # pragma: no cover
    from map import Map
    from system_operations import write_to_log


class LandscapeMetrics(Map):
    """
    Calculates the mean nearest-neighbour for cells across a landscape. See :ref:`here <landscape_metrics>` for details.
    """

    def __init__(self, file=None, logging_level=logging.WARNING):
        """
        Initialises the loggers and default map variables.

        :param file: the path to the map file
        :param logging_level: the logging level to report at
        """
        Map.__init__(self, file)
        self.logger = logging.Logger("necsim")
        self._create_logger(logging_level=logging_level)
        self.c_LM_calc = CLandscapeMetricsCalculator(self.logger)

    def __del__(self):
        """Safely destroy the C++ objects."""
        self.c_LM_calc = None
        Map.__del__(self)

    def get_mnn(self):
        """
        Calculates the mean nearest-neighbour for cells across a landscape. See :ref:`here <landscape_metrics_mnn>` for
        details.

        :return: the mean distance to the nearest neighbour of a cell.

        :rtype: float
        """
        self.c_LM_calc.import_map(self.file_name)
        return self.c_LM_calc.calculate_MNN()

    def get_clumpiness(self):
        """
        Calculates the clumpiness metric for the landscape, a measure of how spread out the points are across the
        landscape. See :ref:`here <landscape_metrics_clumpy>` for details.

        :return: the CLUMPY metric

        :rtype: float
        """
        self.c_LM_calc.import_map(self.file_name)
        return self.c_LM_calc.calculate_CLUMPY()



"""Tests the pause and resume functionality of simulations to ensure it is the same as a continuous simulation."""
import logging
import os
import platform
import shutil
import sqlite3
import unittest

from setup_tests import setUpAll, tearDownAll, skipLongTest

from pycoalescence import Simulation, CoalescenceTree
from pycoalescence.sqlite_connection import check_sql_table_exist
from pycoalescence.future_except import FileNotFoundError, FileExistsError


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()
    for path in ["output spaced", "output spaced2"]:
        if os.path.exists(path):
            shutil.rmtree(path)


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestSimulationPause(unittest.TestCase):
    """
    Test a simple run on a landscape using sampling
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=40)
        self.coal2 = Simulation(logging_level=40)
        self.tree2 = CoalescenceTree()
        self.coal.set_simulation_parameters(
            seed=10,
            task=6,
            output_directory="output",
            min_speciation_rate=0.05,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        self.coal.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal.run()
        self.coal2.set_simulation_parameters(
            seed=10,
            task=7,
            output_directory="output",
            min_speciation_rate=0.05,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        self.coal2.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal2.run()
        self.tree2.set_database(self.coal2)
        self.tree2.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree2.apply()
        self.tree1 = CoalescenceTree()

    def testCanPause(self):
        """
        Tests that simulations can pause execution and correctly store their state to the SQL database (for in-process
        analysis). Checks that the SQL database has correctly written the simulation parameters and that errors are
        thrown when one tries to connect to an incomplete simulation.
        """
        tree2 = CoalescenceTree()
        with self.assertRaises(IOError):
            tree2.set_database(self.coal)
        actual_sim_parameters = dict(
            seed=10,
            task=6,
            output_dir="output",
            speciation_rate=0.05,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=0,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = tree2.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key], msg="Error in {}".format(key))

    def testCanResume(self):
        """
        Tests that simulations can resume execution.
        """
        self.coal.resume_coalescence("output", 10, 6, 10, protracted=False, spatial=True)
        self.tree1.set_database(self.coal)
        actual_sim_parameters = dict(
            seed=10,
            task=6,
            output_dir="output",
            speciation_rate=0.05,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=1,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = self.tree1.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key])

    def testPauseSimMatchesSingleRunSim(self):
        """
        Tests that the two simulations (either pausing, then resuming, or just running straight to completion) produce
        identical results. Checks using comparison of the SPECIES_LIST tables
        """
        self.tree1.set_database(self.coal)
        self.tree1.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree1.apply()
        dict1 = self.tree1.get_simulation_parameters()
        dict2 = self.tree2.get_simulation_parameters()
        for key in dict1.keys():
            if key != "task" and key != "max_time":
                self.assertEqual(dict1[key], dict2[key], "{} not equal.".format(key))
        self.assertEqual(self.coal.get_species_richness(), self.coal2.get_species_richness())
        single_run_species_list = list(self.tree1.get_species_list())
        pause_sim_species_list = list(self.tree2.get_species_list())
        self.assertAlmostEqual(single_run_species_list[0][9], pause_sim_species_list[0][9], 16)
        self.assertListEqual([x for x in pause_sim_species_list], [x for x in single_run_species_list])


class TestSimulationPause2(unittest.TestCase):
    """
    Test a simple run on a landscape using sampling with pause/resume functionality with protractedness
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=logging.CRITICAL)  # TODO set these back
        self.coal2 = Simulation(logging_level=logging.CRITICAL)
        self.tree2 = CoalescenceTree(logging_level=logging.CRITICAL)
        self.coal.set_simulation_parameters(
            seed=10,
            task=26,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            protracted=True,
            min_speciation_gen=0.0,
            max_speciation_gen=100,
        )
        self.coal3 = Simulation(logging_level=logging.CRITICAL)
        self.coal3.set_simulation_parameters(
            seed=10,
            task=26,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            protracted=True,
            min_speciation_gen=0.0,
            max_speciation_gen=100,
        )
        # self.coal.set_simulation_parameters(6, 6, "output", 0.5, 4, 4, 1, 0.1, 1, 1, 200, 0, 200, "null")
        self.coal.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal3.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        # self.coal.detect_map_dimensions()
        self.coal.run()
        try:
            self.coal3.finalise_setup()
        except FileExistsError:
            pass
        self.coal3.run_coalescence()
        self.coal3.apply_speciation_rates()
        self.coal2.set_simulation_parameters(
            seed=10,
            task=27,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            protracted=True,
            min_speciation_gen=0.0,
            max_speciation_gen=100,
        )
        # self.coal.set_simulation_parameters(6, 6, "output", 0.5, 4, 4, 1, 0.1, 1, 1, 200, 0, 200, "null")
        self.coal2.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal2.set_speciation_rates([0.5])
        self.coal2.run()
        self.tree2.set_database(self.coal2)
        self.tree2.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree2.apply()
        self.tree1 = CoalescenceTree()

    def testCanResume2(self):
        """
        Tests that simulations can resume execution by detecting the paused files
        """
        self.tree1.set_database(self.coal3)
        actual_sim_parameters = dict(
            seed=10,
            task=26,
            output_dir="output",
            speciation_rate=0.5,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=1,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=1,
            min_speciation_gen=0.0,
            max_speciation_gen=100.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = self.tree1.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key])

    def testRaisesErrorOnProtractedResume(self):
        """
        Tests that an error is raise if the protracted simulation is attempted to be run from a non-protracted paused
        file, or visa versa.
        """
        coaltmp = Simulation()
        coaltmp.set_simulation_parameters(
            seed=10,
            task=26,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
            protracted=False,
            min_speciation_gen=0.0,
            max_speciation_gen=100,
        )
        with self.assertRaises(RuntimeError):
            coaltmp.resume_coalescence(task=26, seed=10, pause_directory="output", max_time=10, out_directory="output")

    def testPauseSimMatchesSingleRunSim2(self):
        """
        Tests that the two simulations (either pausing, then resuming, or just running straight to completion) produce
        identical results. Checks using comparison of the SPECIES_LIST tables
        """
        self.tree1.set_database(self.coal3)
        self.tree1.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree1.apply()
        self.assertEqual(self.coal3.get_species_richness(), self.coal2.get_species_richness())
        single_run_species_list = list(self.tree1.get_species_list())
        pause_sim_species_list = list(self.tree2.get_species_list())
        dict1 = self.tree1.get_simulation_parameters()
        dict2 = self.tree2.get_simulation_parameters()
        for key in dict1.keys():
            if key != "task":
                self.assertEqual(dict1[key], dict2[key], "{} not equal.".format(key))
        # print(pause_sim_species_list)
        # print(single_run_species_list)
        self.assertAlmostEqual(single_run_species_list[0][9], pause_sim_species_list[0][9], 16)
        self.assertListEqual([x for x in pause_sim_species_list], [x for x in single_run_species_list])


class TestSimulationPause3(unittest.TestCase):
    """
    Test a simple run on a landscape using sampling, moving the paused files between simulations.
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=40)
        self.coal2 = Simulation(logging_level=40)
        self.tree2 = CoalescenceTree()
        self.coal.set_simulation_parameters(
            seed=10,
            task=16,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        # self.coal.set_simulation_parameters(6, 6, "output", 0.5, 4, 4, 1, 0.1, 1, 1, 200, 0, 200, "null")
        self.coal.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        # self.coal.detect_map_dimensions()
        self.coal.run()
        self.coal2.set_simulation_parameters(
            seed=10,
            task=17,
            output_directory="output",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        # self.coal.set_simulation_parameters(6, 6, "output", 0.5, 4, 4, 1, 0.1, 1, 1, 200, 0, 200, "null")
        self.coal2.set_map_files(
            sample_file="sample/SA_samplemaskINT.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal2.set_speciation_rates([0.5])
        self.coal2.run()
        self.tree2.set_database(self.coal2)
        self.tree2.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree2.apply()
        self.tree1 = CoalescenceTree()

    @classmethod
    def tearDownClass(self):
        shutil.rmtree("output2", ignore_errors=True)
        shutil.rmtree("output3", ignore_errors=True)

    def testCanPause(self):
        """
        Tests that simulations can pause executation and correctly store their state to the SQL database (for in-process
        analysis). Checks that the SQL database has correctly written the simulation parameters and that errors are
        thrown when one tries to connect to an incomplete simulation.
        """
        tree2 = CoalescenceTree()
        with self.assertRaises(IOError):
            tree2.set_database(self.coal)
        actual_sim_parameters = dict(
            seed=10,
            task=16,
            output_dir="output",
            speciation_rate=0.5,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=0,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = tree2.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key], msg="Error in {}".format(key))

    def testCanResume(self):
        """
        Tests that simulations can resume execution.
        """
        paused_file_list = ["Dump_main_"]
        os.mkdir("output2")
        os.mkdir("output2/Pause")
        for file in paused_file_list:
            os.rename(
                os.path.join("output", "Pause", file + "16_10.csv"),
                os.path.join("output2", "Pause", file + "16_10.csv"),
            )
        if not os.path.exists("output3"):
            os.mkdir("output3")
        self.coal.resume_coalescence("output2", 10, 16, 10, out_directory="output3")
        self.tree1.set_database(self.coal)
        actual_sim_parameters = dict(
            seed=10,
            task=16,
            output_dir="output3",
            speciation_rate=0.5,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=1,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = self.tree1.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key])

    def testPauseSimMatchesSingleRunSim(self):
        """
        Tests that the two simulations (either pausing, then resuming, or just running straight to completion) produce
        identical results. Checks using comparison of the SPECIES_LIST tables
        """
        self.tree1.set_database(self.coal)
        self.tree1.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree1.apply()
        self.assertEqual(self.coal.get_species_richness(), self.coal2.get_species_richness())
        dict1 = self.tree1.get_simulation_parameters()
        dict2 = self.tree2.get_simulation_parameters()
        for key in dict1.keys():
            if key != "task" and key != "max_time" and key != "output_dir":
                self.assertEqual(dict1[key], dict2[key], "{} not equal.".format(key))
        single_run_species_list = list(self.tree1.get_species_list())
        pause_sim_species_list = list(self.tree2.get_species_list())
        self.assertAlmostEqual(single_run_species_list[0][9], pause_sim_species_list[0][9], 16)
        self.assertListEqual([x for x in pause_sim_species_list], [x for x in single_run_species_list])


@skipLongTest
class TestSimulationPause4(unittest.TestCase):
    """
    Test a simple run on a landscape using sampling, moving the paused files between simulations.
    """

    @classmethod
    def setUpClass(self):
        """
        Sets up the Coalescence object test case.
        """
        self.coal = Simulation(logging_level=logging.CRITICAL)
        self.coal2 = Simulation(logging_level=logging.CRITICAL)
        self.tree2 = CoalescenceTree()
        self.coal.set_simulation_parameters(
            seed=11,
            task=16,
            output_directory="output spaced",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        # self.coal.set_simulation_parameters(6, 6, "output", 0.5, 4, 4, 1, 0.1, 1, 1, 200, 0, 200, "null")
        self.coal.set_map_files(
            sample_file="sample/SA_samplemaskINT spaced.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        # self.coal.detect_map_dimensions()
        self.coal.run()
        self.coal2.set_simulation_parameters(
            seed=11,
            task=17,
            output_directory="output spaced",
            min_speciation_rate=0.5,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        self.coal2.set_map_files(
            sample_file="sample/SA_samplemaskINT spaced.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        self.coal2.set_speciation_rates([0.5])
        self.coal2.run()
        self.tree2.set_database(self.coal2)
        self.tree2.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree2.apply()

        # self.tree2.calculate_octaves()
        # self.tree2.calculate_richness()
        self.tree3 = CoalescenceTree()
        self.coal3 = Simulation(logging_level=logging.CRITICAL)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree("output2 spaced", ignore_errors=True)
        shutil.rmtree("output spaced", ignore_errors=True)

    def testCanPause(self):
        """
        Tests that simulations can pause executation and correctly store their state to the SQL database (for in-process
        analysis). Checks that the SQL database has correctly written the simulation parameters and that errors are
        thrown when one tries to connect to an incomplete simulation.
        """
        tree1 = CoalescenceTree()
        with self.assertRaises(IOError):
            tree1.set_database(self.coal)
        actual_sim_parameters = dict(
            seed=11,
            task=16,
            output_dir="output spaced",
            speciation_rate=0.5,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT spaced.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=0,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = tree1.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key], msg="Error in {}".format(key))

    def testCanResume(self):
        """
        Tests that simulations can resume execution.
        """
        if not os.path.exists("output2 spaced"):
            os.mkdir("output2 spaced")
        self.coal3.resume_coalescence(
            pause_directory="output spaced", seed=11, task=16, max_time=10, out_directory="output2 spaced"
        )
        self.tree3.set_database(self.coal3)
        actual_sim_parameters = dict(
            seed=11,
            task=16,
            output_dir="output2 spaced",
            speciation_rate=0.5,
            sigma=2.0,
            tau=2.0,
            deme=1,
            sample_size=0.1,
            max_time=10,
            dispersal_relative_cost=1.0,
            min_num_species=1,
            habitat_change_rate=0.0,
            gen_since_historical=0.0,
            coarse_map_file="sample/SA_sample_coarse.tif",
            coarse_map_x=35,
            coarse_map_y=41,
            coarse_map_x_offset=11,
            coarse_map_y_offset=14,
            coarse_map_scale=1.0,
            fine_map_file="sample/SA_sample_fine.tif",
            fine_map_x=13,
            fine_map_y=13,
            fine_map_x_offset=0,
            fine_map_y_offset=0,
            sample_file="sample/SA_samplemaskINT spaced.tif",
            grid_x=13,
            grid_y=13,
            sample_x=13,
            sample_y=13,
            sample_x_offset=0,
            sample_y_offset=0,
            historical_coarse_map="none",
            historical_fine_map="none",
            sim_complete=1,
            dispersal_method="normal",
            m_probability=0.0,
            cutoff=0.0,
            landscape_type="closed",
            protracted=0,
            min_speciation_gen=0.0,
            max_speciation_gen=0.0,
            dispersal_map="none",
            time_config_file="null",
        )
        params = self.tree3.get_simulation_parameters()
        for key in params.keys():
            self.assertEqual(params[key], actual_sim_parameters[key])

    def testPauseSimMatchesSingleRunSim(self):
        """
        Tests that the two simulations (either pausing, then resuming, or just running straight to completion) produce
        identical results. Checks using comparison of the SPECIES_LIST tables
        """
        self.tree3.set_database(self.coal3)
        self.tree3.set_speciation_parameters(
            speciation_rates=[0.6, 0.7], record_spatial="T", record_fragments="F", sample_file="null"
        )
        self.tree3.apply()
        self.assertEqual(self.tree3.get_species_richness(), self.tree2.get_species_richness())
        dict1 = self.tree3.get_simulation_parameters()
        dict2 = self.tree2.get_simulation_parameters()
        for key in dict1.keys():
            if key != "task" and key != "max_time" and key != "output_dir":
                self.assertEqual(dict1[key], dict2[key], "{} not equal.".format(key))
        single_run_species_list = list(self.tree3.get_species_list())
        pause_sim_species_list = list(self.tree2.get_species_list())
        # print(pause_sim_species_list)
        # print(single_run_species_list)
        self.assertAlmostEqual(single_run_species_list[0][9], pause_sim_species_list[0][9], 16)
        self.assertListEqual([x for x in pause_sim_species_list], [x for x in single_run_species_list])


class TestPauseSpeciateRemaining(unittest.TestCase):
    """Tests functionality relating to termination of a paused simulation by speciating every remaining lineage."""

    @classmethod
    def setUpClass(cls):
        """Runs a simulation to pause and restart from."""
        cls.sim = Simulation(logging_level=50)
        cls.sim.set_simulation_parameters(
            seed=14,
            task=17,
            output_directory="output",
            min_speciation_rate=0.00000000001,
            sigma=2,
            tau=2,
            deme=1,
            sample_size=0.1,
            max_time=0,
            dispersal_relative_cost=1,
            min_num_species=1,
            dispersal_method="normal",
        )
        cls.sim.set_map_files(
            sample_file="sample/SA_samplemaskINT spaced.tif",
            fine_file="sample/SA_sample_fine.tif",
            coarse_file="sample/SA_sample_coarse.tif",
        )
        cls.sim.run()
        shutil.copy(cls.sim.output_database, os.path.join("output", "data_17_14b.db"))
        cls.ct = CoalescenceTree()
        cls.ct.speciate_remaining(cls.sim)

    def testPausedSimulationExists(self):
        """Tests that the paused simulation exists and the tables have been stored properly."""
        paused_db = os.path.join("output", "data_{}_{}b.db".format(17, 14))
        self.assertTrue(os.path.exists(paused_db))
        with self.assertRaises(IOError):
            ct = CoalescenceTree(paused_db)
        ct = None
        database = sqlite3.connect(paused_db)
        self.assertTrue(check_sql_table_exist(database, "SPECIES_LIST"))
        self.assertTrue(check_sql_table_exist(database, "SIMULATION_PARAMETERS"))
        self.assertFalse(
            bool(database.cursor().execute("SELECT sim_complete FROM SIMULATION_PARAMETERS").fetchall()[0][0])
        )

    @unittest.skipIf(platform.system() == "Windows", "Skipping tests not compatible with Windows.")
    def testForcedSpeciation(self):
        """Tests that the forced speciation of remaining lineages works as expected."""
        ct = CoalescenceTree(self.sim)
        ct.set_speciation_parameters(speciation_rates=[0.1, 0.5, 0.95])
        ct.apply()
        # if "win" in sys.platform:
        # 	self.assertEqual(1171, ct.get_species_richness(1))
        # 	self.assertEqual(1172, ct.get_species_richness(2))
        # 	self.assertEqual(1173, ct.get_species_richness(3))
        # else:
        self.assertEqual(1170, ct.get_species_richness(1))
        self.assertEqual(1170, ct.get_species_richness(2))
        self.assertEqual(1173, ct.get_species_richness(3))
        ct = CoalescenceTree()
        with self.assertRaises(FileNotFoundError):
            ct.speciate_remaining("notafile.db")
        with self.assertRaises(IOError):
            ct.speciate_remaining(self.sim)



import logging
import math
import os
import unittest

from setup_tests import setUpAll, tearDownAll, skipLongTest

from pycoalescence import Map
from pycoalescence.dispersal_simulation import DispersalSimulation


def setUpModule():
    """
    Creates the output directory and moves logging files
    """
    setUpAll()


def tearDownModule():
    """
    Removes the output directory
    """
    tearDownAll()


class TestDispersalSimulation(unittest.TestCase):
    """
    Tests the dispersal simulation methods contained in Map
    """

    @classmethod
    def setUpClass(cls):
        """
        Sets up the class by running the dispersal simulations for later reference.
        """
        cls.m = DispersalSimulation(logging_level=logging.CRITICAL)
        cls.m.set_map_files(os.path.join("sample", "SA_sample_fine.tif"))
        cls.m.set_simulation_parameters(
            number_repeats=100,
            output_database=os.path.join("output", "sim_pars_test4.db"),
            seed=2,
            sigma=2,
            landscape_type="tiled_fine",
        )
        cls.m.run_mean_dispersal()
        cls.m.set_simulation_parameters(number_repeats=100, seed=2, sigma=2, landscape_type="tiled_fine")
        cls.m.run_mean_dispersal()
        cls.m.set_simulation_parameters(
            number_repeats=100, seed=4, sigma=3, tau=1, landscape_type="tiled_fine", dispersal_method="fat-tail"
        )
        cls.m.run_mean_dispersal()
        cls.m.set_simulation_parameters(
            number_repeats=100, number_steps=10, seed=2, sigma=2, landscape_type="tiled_fine"
        )
        cls.m.run_mean_distance_travelled()
        cls.m.set_simulation_parameters(
            number_repeats=200, number_steps=10, seed=2, sigma=5, landscape_type="tiled_fine"
        )
        cls.m.run_mean_distance_travelled()
        cls.m.set_simulation_parameters(
            number_repeats=100, number_steps=20, seed=5, sigma=2, landscape_type="tiled_fine"
        )
        cls.m.run_mean_distance_travelled()

    def testGetFromDatabaseErrors(self):
        """Tests that the correct errors are raised when obtaining from a database."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.dispersal_database = "doesnotexist.db"
        with self.assertRaises(IOError):
            m.get_mean_dispersal()
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_mean_dispersal(database=os.path.join("sample", "sample.db"))
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_mean_distance_travelled(database=os.path.join("sample", "sample.db"))
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_stdev_dispersal(database=os.path.join("sample", "sample.db"))
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_stdev_distance_travelled(database=os.path.join("sample", "sample.db"))
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_all_distances(database=os.path.join("sample", "sample.db"))
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(IOError):
            m.get_all_dispersal(database=os.path.join("sample", "sample.db"))

    def testRaisesValueErrorNullNotSet(self):
        """
        Tests that a ValueError is raised when the map file is "null" but dimensions have not been manually set.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(ValueError):
            m.set_simulation_parameters(number_repeats=10000, output_database="output/normaldispersal.db", seed=1)
            m.set_map_files("null")

    def testRaisesError(self):
        """
        Tests that a dispersal.Error is raised when incorrect dispersal method is provided.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        with self.assertRaises(RuntimeError):
            m.set_simulation_parameters(
                number_repeats=10000, output_database="output/emptydb.db", seed=1, dispersal_method="notamethod"
            )
            m.set_map_files("sample/SA_sample_fine.tif")
            try:
                m.run_mean_dispersal()
            except Exception as e:
                raise e

    def testDispersalNullOutputs(self):
        """
        Tests that the dispersal simulations accurately generate the correct distance distribution for null landscapes.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("null", 10, 10)
        m.set_simulation_parameters(
            number_repeats=100000,
            output_database="output/normaldispersal1.db",
            seed=2,
            sigma=10,
            landscape_type="infinite",
            restrict_self=False,
        )
        m.run_mean_dispersal()
        self.assertAlmostEqual(m.get_mean_dispersal(), 10 * (3.14 ** 0.5) / 2 ** 0.5, places=2)
        expected_dispersal = [
            3.1622776601683795,
            27.202941017470888,
            16.1245154965971,
            15.132745950421556,
            13.038404810405298,
            13.601470508735444,
            11.180339887498949,
            13.038404810405298,
            8.48528137423857,
            12.083045973594572,
        ]

        actual_dispersal = m.get_all_dispersal()[:10]
        for expected, actual in zip(expected_dispersal, actual_dispersal):
            self.assertAlmostEqual(expected, actual, places=3)

    @skipLongTest
    def testLandscapeTypesMatch(self):
        """
        Tests that a null landscape with tiled_fine matches a null landscape with infinite
        """
        m1 = DispersalSimulation(logging_level=logging.CRITICAL)
        m1.set_map("null", 10, 10)
        m1.set_simulation_parameters(
            number_repeats=100000,
            output_database="output/normaldispersal2.db",
            seed=2,
            sigma=2,
            landscape_type="infinite",
            restrict_self=False,
        )
        m1.run_mean_dispersal()
        m2 = DispersalSimulation(logging_level=logging.CRITICAL)
        m2.set_map("null", 10, 10)
        m2.set_simulation_parameters(
            number_repeats=100000,
            output_database="output/normaldispersal3.db",
            seed=2,
            sigma=2,
            landscape_type="tiled_fine",
            restrict_self=False,
        )
        m2.run_mean_dispersal()
        self.assertEqual(m2.get_mean_dispersal(), m1.get_mean_dispersal())

    def testDispersalMapOutputs(self):
        """
        Tests that the dispersal simulations work as intended for real maps.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=10000,
            output_database="output/realdispersal.db",
            seed=1,
            sigma=1,
            landscape_type="tiled_fine",
        )
        m.run_mean_dispersal()
        self.assertAlmostEqual(1.2641113312060646, m.get_mean_dispersal(), places=4)

    def testDispersalDistanceTravelled(self):
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=100,
            number_steps=10,
            output_database="output/sim_pars_test2.db",
            seed=2,
            sigma=1,
            landscape_type="tiled_fine",
        )
        m.run_mean_distance_travelled()
        self.assertAlmostEqual(4.206153698181416, m.get_mean_distance_travelled(), places=3)
        expected_distances = [
            4.123105625617661,
            6.4031242374328485,
            6.0,
            5.0,
            5.0,
            1.4142135623730951,
            2.23606797749979,
            2.23606797749979,
            2.8284271247461903,
            3.1622776601683795,
        ]

        actual_dispersal = m.get_all_distances()[:10]
        for expected, actual in zip(expected_distances, actual_dispersal):
            self.assertAlmostEqual(expected, actual, places=3)

    def testDispersalSimulationParametersStoredCorrectly(self):
        """
        Tests that simulation parameters are stored correctly in the database.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=100, output_database="output/sim_pars_test1.db", seed=1, sigma=1, landscape_type="tiled_fine"
        )
        m.run_mean_dispersal()
        m.update_parameters(sigma=2, number_steps=10, seed=2)
        m.run_mean_distance_travelled()
        main_dict = m.get_database_parameters()
        comparison_dict = {
            1: {
                "simulation_type": "DISPERSAL_DISTANCES",
                "sigma": 1,
                "tau": 1,
                "cutoff": 100,
                "m_prob": 1.0,
                "dispersal_method": "normal",
                "map_file": "sample/SA_sample_fine.tif",
                "seed": 1,
                "number_steps": 1,
                "number_repeats": 100,
            },
            2: {
                "simulation_type": "DISTANCES_TRAVELLED",
                "sigma": 2,
                "tau": 1,
                "cutoff": 100.0,
                "m_prob": 1.0,
                "dispersal_method": "normal",
                "map_file": "sample/SA_sample_fine.tif",
                "seed": 2,
                "number_steps": 10,
                "number_repeats": 100,
            },
        }
        for key in main_dict.keys():
            for inner_key in main_dict[key].keys():
                self.assertEqual(
                    main_dict[key][inner_key],
                    comparison_dict[key][inner_key],
                    msg="Error in {}, {} != {}".format(
                        inner_key, main_dict[key][inner_key], comparison_dict[key][inner_key]
                    ),
                )

    def testMultipleSimulationsAverages(self):
        """
        Tests that with multiple simulations, the averages are outputted correctly.
        """

        self.assertAlmostEqual(2.5016203799003596, self.m.get_mean_dispersal(parameter_reference=2))
        self.assertAlmostEqual(17.448853121157676, self.m.get_mean_dispersal(parameter_reference=3))
        self.assertAlmostEqual(7.7522237785164885, self.m.get_mean_distance_travelled(parameter_reference=4))
        self.assertAlmostEqual(19.15984097911993, self.m.get_mean_distance_travelled(parameter_reference=5))

    def testStandardDeviationDistances(self):
        """
        Tests that the standard deviation is correctly returned from the simulated database
        """
        self.assertAlmostEqual(1.2537525572724404, self.m.get_stdev_dispersal(parameter_reference=1))
        self.assertAlmostEqual(1.2537525572724404, self.m.get_stdev_dispersal(parameter_reference=2))
        self.assertAlmostEqual(4.545660181734175, self.m.get_stdev_distance_travelled(parameter_reference=4))
        self.assertAlmostEqual(11.1072270911707, self.m.get_stdev_distance_travelled(parameter_reference=5))

    def testDatabaseReferences(self):
        """
        Tests that the database parameters are stored and returned correctly.
        """
        self.assertEqual([x for x in range(1, 7)], self.m.get_database_references())
        expected_dict = {
            "simulation_type": "DISPERSAL_DISTANCES",
            "sigma": 2,
            "tau": 1,
            "m_prob": 1.0,
            "cutoff": 100.0,
            "dispersal_method": "normal",
            "map_file": os.path.join("sample", "SA_sample_fine.tif"),
            "seed": 2,
            "number_steps": 1,
            "number_repeats": 100,
        }
        self.assertEqual(expected_dict, self.m.get_database_parameters()[1])
        self.assertEqual(expected_dict, self.m.get_database_parameters(1))
        for key in range(1, 6):
            self.m.get_database_parameters(key)
        with self.assertRaises(KeyError):
            self.m.get_database_parameters(7)
        with self.assertRaises(KeyError):
            self.m.get_database_parameters("NotAKey")

    def testDispersalMapReading(self):
        """
        Tests that dispersal simulations are correctly read from a completed simulation
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=10000,
            output_database="output/realdispersal2.db",
            seed=1,
            sigma=1,
            landscape_type="tiled_fine",
        )
        m.run_mean_dispersal()
        m2 = DispersalSimulation(dispersal_db=m.dispersal_database, logging_level=logging.CRITICAL)
        self.assertAlmostEqual(1.2641113312060646, m2.get_mean_dispersal(), places=4)
        m3 = DispersalSimulation(dispersal_db=m, logging_level=logging.CRITICAL)
        self.assertAlmostEqual(1.2641113312060646, m3.get_mean_dispersal(), places=4)

    def testAlternativeMapFileSetting(self):
        m = DispersalSimulation(file="sample/SA_sample_fine.tif", logging_level=logging.CRITICAL)
        if m.fine_map.file_name is None:
            self.fail("File name not correctly set.")
        m.set_simulation_parameters(
            number_steps=10,
            number_repeats=10,
            output_database="output/realdispersal3.db",
            sigma=1,
            landscape_type="tiled_fine",
            seed=1,
        )
        m.run_mean_distance_travelled()
        m.run_mean_dispersal()
        with self.assertRaises(ValueError):
            m.get_mean_dispersal(parameter_reference=3)
        with self.assertRaises(ValueError):
            m.get_all_dispersal(parameter_reference=3)
        with self.assertRaises(ValueError):
            m.get_mean_distance_travelled(parameter_reference=3)
        with self.assertRaises(ValueError):
            m.get_all_distances(parameter_reference=3)
        self.assertAlmostEqual(1.2892922226992172, m.get_mean_dispersal(parameter_reference=2), places=4)

    def testParameterUpdating(self):
        """
        Tests that a simulation can be run properly with updating parameters in between.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map("sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=100, output_database="output/realdispersal4.db", seed=1, sigma=1, landscape_type="tiled_fine"
        )
        m.run_mean_dispersal()
        m.update_parameters(number_repeats=6, sigma=10)
        m.run_mean_dispersal()
        m.update_parameters(dispersal_method="fat-tail", number_steps=10)
        m.run_mean_distance_travelled()
        expected_params = {
            1: {
                "simulation_type": "DISPERSAL_DISTANCES",
                "sigma": 1.0,
                "tau": 1.0,
                "m_prob": 1.0,
                "cutoff": 100.0,
                "dispersal_method": "normal",
                "map_file": "sample/SA_sample_fine.tif",
                "seed": 1,
                "number_steps": 1,
                "number_repeats": 100,
            },
            2: {
                "simulation_type": "DISPERSAL_DISTANCES",
                "sigma": 10.0,
                "tau": 1.0,
                "m_prob": 1.0,
                "cutoff": 100.0,
                "dispersal_method": "normal",
                "map_file": "sample/SA_sample_fine.tif",
                "seed": 1,
                "number_steps": 1,
                "number_repeats": 6,
            },
            3: {
                "simulation_type": "DISTANCES_TRAVELLED",
                "sigma": 10.0,
                "tau": 1.0,
                "m_prob": 1.0,
                "cutoff": 100.0,
                "dispersal_method": "fat-tail",
                "map_file": "sample/SA_sample_fine.tif",
                "seed": 1,
                "number_steps": 10,
                "number_repeats": 6,
            },
        }
        actual_params = m.get_database_parameters()
        for key in expected_params.keys():
            self.assertEqual(expected_params[key], actual_params[key])
        self.assertAlmostEqual(1.245882030374995, m.get_mean_dispersal(parameter_reference=1), places=4)
        self.assertAlmostEqual(12.035293581998125, m.get_mean_dispersal(parameter_reference=2), places=4)
        self.assertAlmostEqual(507.88676714308644, m.get_mean_distance_travelled(parameter_reference=3), places=4)

    def testDispersalWithCoarse(self):
        """Tests a dispersal simulation using a coarse and fine map."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/SA_sample_fine.tif", coarse_file="sample/SA_sample_coarse.tif")
        m.set_simulation_parameters(
            number_repeats=100, output_database="output/realdispersal5.db", seed=1, sigma=1, landscape_type="tiled_fine"
        )
        m.run_mean_dispersal()
        self.assertAlmostEqual(1.2964612543012595, m.get_mean_dispersal(parameter_reference=1), places=4)

    def testDispersalWithSample(self):
        """Tests a dispersal simulation using a sample and fine map."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/SA_fine_expanded.tif", sample_file="sample/SA_sample_fine.tif")
        m.set_simulation_parameters(
            number_repeats=100, output_database="output/realdispersal6.db", seed=1, sigma=1, landscape_type="closed"
        )
        m.run_mean_dispersal()
        self.assertAlmostEqual(0.9840668874233433, m.get_mean_dispersal(parameter_reference=1), places=4)

    def testApplyingMultipleNumberSteps(self):
        """Tests running a single simulation with multiple number of steps at once."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/SA_sample_fine.tif", coarse_file="sample/SA_sample_coarse.tif")
        m.set_simulation_parameters(
            number_repeats=100, output_database="output/realdispersal7.db", seed=1, sigma=1, landscape_type="tiled_fine"
        )
        m.run_mean_distance_travelled(number_steps=[10, 20, 30], number_repeats=100, seed=1)
        m.update_parameters(number_steps=[40, 50, 60])
        m.run_mean_distance_travelled()
        for k, v in [(1, 10), (2, 20), (3, 30), (4, 40), (5, 50), (6, 60)]:
            self.assertEqual(v, m.get_database_parameters()[k]["number_steps"])
        self.assertAlmostEqual(4.001297548429019, m.get_mean_distance_travelled(parameter_reference=1), places=6)
        self.assertAlmostEqual(6.1136500795111886, m.get_mean_distance_travelled(parameter_reference=2), places=6)
        self.assertAlmostEqual(7.3630228316363295, m.get_mean_distance_travelled(parameter_reference=3), places=6)
        self.assertAlmostEqual(8.497107432997838, m.get_mean_distance_travelled(parameter_reference=4), places=6)
        self.assertAlmostEqual(9.723744156016298, m.get_mean_distance_travelled(parameter_reference=5), places=6)
        self.assertAlmostEqual(10.81902521181573, m.get_mean_distance_travelled(parameter_reference=6), places=6)

    def testRunMeanDistanceErrors(self):
        """Tests that the mean distance travelled errors are correctly raised."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/SA_sample_fine.tif", coarse_file="sample/SA_sample_coarse.tif")
        with self.assertRaises(ValueError):
            m.run_mean_distance_travelled()
        with self.assertRaises(ValueError):
            m.run_mean_distance_travelled(number_repeats=1)
        with self.assertRaises(ValueError):
            m.run_mean_distance_travelled(number_steps=1, number_repeats=1)

    def testRunMeanDispersalErrors(self):
        """Tests that the mean dispersal errors are correctly raised."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/SA_sample_fine.tif", coarse_file="sample/SA_sample_coarse.tif")
        with self.assertRaises(ValueError):
            m.run_mean_dispersal()

    @skipLongTest
    def testNullDispersalWithCoarse(self):
        """Sanity checks that the distances are calculated properly on a coarse map."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_map_files(fine_file="sample/null.tif", coarse_file="sample/null_large.tif")
        m.set_simulation_parameters(
            number_repeats=100000,
            output_database="output/realdispersal8.db",
            seed=2,
            sigma=2,
            landscape_type="closed",
            number_steps=100,
        )
        m.run_mean_distance_travelled()
        m.update_parameters(sigma=4)
        m.run_mean_distance_travelled()
        m.update_parameters(sigma=8)
        m.run_mean_distance_travelled()
        self.assertAlmostEqual(
            2 * (math.pi * 100 / 2) ** 0.5, m.get_mean_distance_travelled(parameter_reference=1), places=0
        )
        self.assertAlmostEqual(
            4 * (math.pi * 100 / 2) ** 0.5, m.get_mean_distance_travelled(parameter_reference=2), places=0
        )
        self.assertAlmostEqual(
            8 * (math.pi * 100 / 2) ** 0.5, m.get_mean_distance_travelled(parameter_reference=3), places=0
        )

    def testDispersalWithHistoricalMaps(self):
        """Tests that a simulation with a historical map works."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_simulation_parameters(
            number_repeats=2,
            output_database="output/realdispersal9.db",
            seed=2,
            sigma=2,
            landscape_type="closed",
            number_steps=100,
        )
        m1 = Map("sample/SA_sample_fine.tif")
        m.set_map_files(m1)
        m.add_historical_map(fine_file="sample/SA_sample_fine2.tif", coarse_file="none", time=10, rate=0.1)
        m.run_mean_distance_travelled()
        self.assertAlmostEqual(4.3471617258268065, m.get_mean_distance_travelled(parameter_reference=1), places=6)

    def testDispersalDatabaseRemoval(self):
        """Tests that the dispersal database connection can be deleted properly, with simulation variables preserved."""
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_simulation_parameters(
            number_repeats=2,
            output_database="output/realdispersal10.db",
            seed=2,
            sigma=2,
            landscape_type="closed",
            number_steps=100,
        )
        m.set_map("sample/SA_sample_fine.tif")
        m.run_mean_distance_travelled()
        mean_dist1 = m.get_mean_distance_travelled()
        self.assertTrue(os.path.exists(m.dispersal_database))
        m._remove_existing_db()
        self.assertFalse(os.path.exists(m.dispersal_database))
        m.run_mean_distance_travelled()
        self.assertTrue(os.path.exists(m.dispersal_database))
        self.assertEqual(mean_dist1, m.get_mean_distance_travelled())

    def testDispersalDatabaseCreation(self):
        """
        Tests that a dispersal database is created initially before the simulation is run, and re-created, if it
        has been deleted, before any simulations.
        """
        m = DispersalSimulation(logging_level=logging.CRITICAL)
        m.set_simulation_parameters(
            number_repeats=2,
            output_database="output/realdispersal11.db",
            seed=2,
            sigma=2,
            landscape_type="closed",
            number_steps=100,
        )
        m.set_map("sample/SA_sample_fine.tif")
        self.assertTrue(os.path.exists(m.dispersal_database))
        m._remove_existing_db()
        self.assertFalse(os.path.exists(m.dispersal_database))
        m.run_mean_distance_travelled()
        self.assertTrue(os.path.exists(m.dispersal_database))
        self.assertAlmostEqual(4.3471617258268065, m.get_mean_distance_travelled(parameter_reference=1), places=6)

    def testRunMeanDistanceTravelledErrors(self):
        """Tests that run mean distance travelled produces the expected errors."""
        d = DispersalSimulation(dispersal_db=os.path.join("output", "output.db"))
        with self.assertRaises(ValueError):
            d.run_mean_distance_travelled()

    def testCheckBaseParametersErrors(self):
        """Tests that the correct errors are raised when the base parameters are checked."""
        d = DispersalSimulation()
        with self.assertRaises(ValueError):
            d.check_base_parameters()
        d.number_repeats = 1
        with self.assertRaises(ValueError):
            d.check_base_parameters()
        d.seed = 1
        with self.assertRaises(ValueError):
            d.check_base_parameters()



"""
Generate fragmented landscapes with specific properties. Detailed :ref:`here <Simulate_landscapes>`.

Contains :class:`.FragmentedLandscape` for creating a fragmented landscape using hexagonal packing and an even spread of
individuals between fragments. Requires scipy and matplotlib.
"""

import copy
import logging
import math
import os

import numpy as np

try:
    from matplotlib.pyplot import figure
    from matplotlib import pyplot as plt
except (ImportError, RuntimeError) as ie:  # pragma: no cover
    logging.warning(ie)
try:
    from osgeo import gdal
except ImportError as ie:  # pragma: no cover
    try:
        import gdal
    except ImportError:
        raise ie
from .spatial_algorithms import lloyds_algorithm, archimedes_spiral
from .system_operations import check_parent
from .future_except import FileExistsError


class Fragment:
    """
    Simple class containing the centres of fragments for a fragmented landscape
    """

    def __init__(self, x=None, y=None):
        """
        Default initiator

        :param x: the x position of the fragment centre
        :param y: the y position of the fragment centre

        :rtype: None
        """
        self.x = 0
        self.y = 0
        self.size = 0
        self.counter = 0
        if x and y:
            self.setup(x, y)

    def setup(self, x, y):
        """
        Sets up the fragment from the x and y position.

        :param x: the x position of the fragment centre
        :param y: the y position of the fragment centre

        :rtype: None
        """
        self.x = copy.copy(x)
        self.y = copy.copy(y)

    def place_on_grid(self):
        """
        Changes the x and y positions to integers (always rounds down).

        :rtype: None
        """
        self.x = int(math.floor(self.x))
        self.y = int(math.floor(self.y))


class FragmentedLandscape:
    """
    Contains hexagonal packing algorithms for spacing clumps evenly on the landscape. Includes a LLoyd's smoothing
    algorithm for better spacing of fragments.

    .. note::

        Fragments will not be distinct units for unfragmented landscapes (with above around 50% habitat cover).

    """

    def __init__(self, number_fragments=None, size=None, total=None, output_file=None):
        """
        Initiates the FragmentLandscape object with the supplied parameters.

        :param number_fragments: the number of individual fragments to exist on the landscape
        :param size: the size of the x and y dimensions of the landscape
        :param total: the total number of individuals to place on the landscape
        :param output_file: the output tif file to write the output to

        :rtype: None
        """
        self.number_fragments = 0
        self.size = 0
        self.output_file = None
        # upper bound for the number of x, y dimensions required
        self.number_rows = 0
        self.number_cols = 0
        self.is_shrunk = False  # this will be true if this height is less than the width
        self.total_added = 0
        self.remainder = 0  # if this number is 0, we have no more extra points to add
        self.needs_voronoi = False
        self.total = 0
        self.grid = None
        self.fragments = None
        if number_fragments and size and total and output_file:
            self.setup(number_fragments, size, total, output_file)

    def setup(self, number_fragments, size, total, output_file):
        """
        Sets up the landscape by checking parameters and setting object sizes.

        :param number_fragments: the number of individual fragments to exist on the landscape
        :param size: the size of the x and y dimensions of the landscape
        :param total: the total number of individuals to place on the landscape
        :param output_file: the output tif file to write the output to

        :rtype: None
        """
        self.number_fragments = number_fragments
        self.size = size
        self.total = total
        self.output_file = output_file
        self.grid = np.zeros(shape=[size, size], dtype=np.int)
        self.fragments = [None for _ in range(self.number_fragments)]
        self._precursory_validation(number_fragments, size, total)

    def generate(self, override_smoothing=None, n=10):
        """
        Convenience function for creating fragments in one function. Generates the landscape and writes out to the
        output file.

        If smoothing is true, will run Lloyd's algorithm
        after the hexagonal packing algorithm to increase the equality of the spacing.

        .. note:: smoothing is recommended for any landscape that is doesn't contain a square number of fragments.

        :param override_smoothing: if true, overrides the default smoothing settings (enabled for landscapes with fewer
            than 100000 fragments.
        :param n: the number of iterations to run Lloyd's algorithm for

        :rtype: None
        """
        self._check_file_and_path()
        self.create(override_smoothing=override_smoothing, n=n)
        self.write_to_raster()

    def create(self, override_smoothing=None, n=10):
        """
        Creates the landscape, including running the hexagonal packing and smoothing algorithms (if required).

        .. note:: smoothing is recommended for any landscape that is doesn't contain a square number of fragments.

        :param override_smoothing: if true, overrides the default smoothing settings (enabled for landscapes with fewer
            than 100000 fragments.

        :param n: the number of iterations to run Lloyd's algorithm for

        :rtype: None
        """
        smoothing = self.number_fragments > 10000
        if override_smoothing is not None:
            smoothing = override_smoothing
        self.place_fragments(smoothing=smoothing, n=n)
        self.fill_grid()

    def _check_file_and_path(self):
        """
        Checks that the file does not already exist and the file path exists.

        :raises FileExistsError: if the file already exists.

        :raises IOError: if the output file is None

        :param file: the file to check for

        :rtype: None
        """
        if self.output_file is None:
            raise IOError("File is None")
        if os.path.exists(self.output_file):
            raise FileExistsError("File already exists at {}.".format(self.output_file))
        check_parent(self.output_file)

    def place_fragments(self, smoothing=True, n=10):
        """
        Places the fragments evenly on the landscape. If smoothing is true, will run Lloyd's algorithm after the
        hexagonal packing algorithm to increase the equality of the spacing.

        .. note:: smoothing is recommended for any landscape that is doesn't contain a square number of fragments.

        :param smoothing: if true, runs Lloyd's algorithm after the hexagonal packing
        :param n: the number of iterations to run Lloyd's algorithm for

        :rtype: None
        """
        self._calculate_dimensions()
        self._place_initial_grid()
        self._add_remainders()
        self._scale_fragment_centres()
        if smoothing:
            self._smooth_points(n=n)

    def fill_grid(self):
        """
        Distributes the sizes evenly between the fragments, generating the actual landscape.

        :rtype: None
        """
        self._fill_fragments()
        for fragment in self.fragments:
            fragment.place_on_grid()
            self._fill_next_cell(fragment.x, fragment.y, 0.1, 0, fragment.size)
        self._final_validation()

    def _fill_fragments(self):
        """
        Generates the sizes for the fragments, splitting the sizes as evenly across the landscape as possible

        :rtype: None
        """
        lower_size = math.floor(self.total / self.number_fragments)
        running_total = 0
        for fragment in self.fragments:
            fragment.size = lower_size
            running_total += lower_size
        i = 0
        diff = self.total - running_total
        if diff != 0:
            increment = self.number_fragments / diff
            while running_total < self.total:
                self.fragments[int(math.floor(i))].size += 1
                i += increment
                if i >= self.number_fragments:
                    i = i % self.number_fragments
                running_total += 1
        running_total = 0
        for frag in self.fragments:
            running_total += frag.size
        if running_total > self.total:  # pragma: no cover
            raise ValueError("Generated total greater than expected total: {} > {}".format(running_total, self.total))

    def _fill_next_cell(self, cell_x, cell_y, radius, theta, size):
        """
        Fills the next cell from the given point by spiralling outwards until it reaches a cell that is empty.

        This uses the archimedian spiral formula from the centre of fragments and calls itself recursively to generate
        the new points.

        :raises TypeError: if the grid is too small for the number of points to add.

        :param cell_x: the x coordinate of the centre of the spiral
        :param cell_y: the y coordinate of the centre of the spiral
        :param radius: the radius for the archimedian spiral. This is increased as cells are checked
        :param theta: the angle for the archimedian spiral. This is increased as cells are checked
        :param size: the number of individuals yet to place

        :rtype: None
        """
        try:
            # Use the archimedian spiral formula
            while size > 0:
                check_x, check_y = archimedes_spiral(cell_x, cell_y, radius, theta)
                check_x = check_x % self.size
                check_y = check_y % self.size
                while self.grid[check_x, check_y] > 0:
                    theta += 0.1 * math.pi / (2 * max(radius, 1))
                    radius = theta / (2 * math.pi)
                    check_x, check_y = archimedes_spiral(cell_x, cell_y, radius, theta)
                    check_x = check_x % self.size
                    check_y = check_y % self.size
                    if radius > self.size:  # pragma: no cover
                        raise TypeError("Grid too small for adding fragment sizes!")
                self.grid[check_x, check_y] = 1
                size -= 1
        except IndexError as ie:  # pragma: no cover
            raise SystemError("Error placing points. Please report this bug: {}".format(ie))

    def _calculate_dimensions(self):
        """
        Calculates the dimensions of the landscape in terms of columns and rows of fragments.

        :rtype: None
        """
        if self.number_fragments == 2:
            self.number_rows = 1
            self.number_cols = 1
            self.remainder = 1
        else:
            self.number_rows = math.ceil(math.sqrt(self.number_fragments))
            if math.pow(self.number_rows, 2) == self.number_fragments:
                # Then the case is very basic. Yay!
                self.number_cols = self.number_rows
            else:
                # Should use Lloyd's algorthim with Voronoi diagrams
                self.number_cols = math.floor(math.sqrt(self.number_fragments))
                self.remainder = self.number_fragments - math.pow(self.number_cols, 2)

    def _place_initial_grid(self):
        """
        Places the initial points on the grid, without worrying about the remainders.

        This is a very basic hexagonal packing algorithm.

        :rtype: None
        """
        for x in range(int(self.number_cols)):
            for y in range(int(self.number_cols)):
                if y % 2 != 0:
                    # offset the odd indices
                    self.fragments[self.total_added] = Fragment(x + 1, y + 0.5)
                else:
                    self.fragments[self.total_added] = Fragment(x + 0.5, y + 0.5)
                self.total_added += 1

    def _add_remainders(self):
        """
        Adds the remainder points to the outside of the grid, adding to odd columns, from the inside out.

        :rtype: None
        """
        if self.total_added + self.remainder != self.number_fragments:
            raise ValueError(
                "Total added ({}) plus remainder ({}) != the number of fragments ({})."
                " Please report this bug.".format(self.total_added, self.remainder, self.number_fragments)
            )
        tot = 0
        if self.remainder > 0:
            iteration = self.number_cols / self.remainder
            while self.remainder > self.number_cols:
                self.fragments[self.total_added] = Fragment(self.number_cols + 0.5, tot)
                self.remainder -= 1
                self.total_added += 1
                tot += iteration
            while self.remainder > 0:
                self.fragments[self.total_added] = Fragment(self.remainder - 0.5, self.number_cols + 0.5)
                self.remainder -= 1
                self.total_added += 1

    def _scale_fragment_centres(self):
        """
        Scales the fragment centres from their values so that they correctly cover the whole landscape.

        :rtype: None
        """
        maximum_x = max([each.x for each in self.fragments]) + 0.5
        maximum_y = max([each.y for each in self.fragments]) + 0.5
        scaling_x = self.size / maximum_x
        scaling_y = self.size / maximum_y
        for fragment in self.fragments:
            fragment.x *= scaling_x
            fragment.y *= scaling_y

    def _smooth_points(self, n=10):
        """
        Smooth the points using Lloyd's algorithm so that they are more evenly spaced. Increase n to increase the number
        of iterations of Lloyd's algorithm and increase the convergence.

        Note that LLoyd's algorithm only finds a local solution for equal distances between points.

        :param n: the number of iterations of Lloyd's algorithm (higher is more even spacing)

        :rtype: None
        """
        points = [[fragment.x, fragment.y] for fragment in self.fragments]
        maxima = (self.size, self.size)
        new_points = lloyds_algorithm(points_list=points, maxima=maxima, n=n)
        for i, fragment in enumerate(self.fragments):
            fragment.x, fragment.y = new_points[i]

    def _precursory_validation(self, number_fragments, size, total):
        """
        Checks that the dimensions and numbers make sense.

        :raises ValueError: if the combination of parameters does not make sense.
        """
        if number_fragments > size * size:
            raise ValueError("Cannot add more fragments than cells on the grid.")
        if number_fragments > total:
            raise ValueError("Cannot add more fragments than total individuals.")
        if total > size * size:
            raise ValueError("Cannot add more individuals than cells on the grid.")

    def _final_validation(self):
        """
        Checks that the number of fragments is correct, that each fragment exists within the plot and

        :raises ValueError: if any of the validation checks fail

        :rtype: None
        """
        try:
            if len(self.fragments) != self.total_added:
                raise ValueError("Fragment number not correct. Please report this bug.")
            for fragment in self.fragments:
                if not 0 <= fragment.x <= self.size or not 0 <= fragment.y <= self.size:
                    raise ValueError("Fragments not set up correctly. Please report this bug.")
            if self.remainder != 0:
                raise ValueError("Remainders not correct. Please report this bug.")
            if np.sum(self.grid) != float(self.total):
                raise ValueError("Total does not equal size: {} != {}.".format(np.sum(self.grid), float(self.total)))
        except ValueError as ve:
            raise ValueError("Validation failed: {}".format(ve))

    def write_to_raster(self):
        """
        Writes the landscape to a tif file.

        :raises FileExistsError: if the output file already exists

        :param output_file: the path to the tif file to write out to.

        :rtype: None
        """
        self._check_file_and_path()
        geotransform = (0, 1, 0, 0, 0, -1)
        # Stupid attempt to fix the no data issue for writing out numpy arrays
        for i in range(self.size):
            if self.grid[i, i] == 0:
                self.grid[i, i] == 0
        output_raster = gdal.GetDriverByName("GTiff").Create(self.output_file, self.size, self.size, 1, gdal.GDT_Byte)
        if not output_raster:  # pragma: no cover
            raise IOError("Could not create tif file at {}.".format(self.output_file))
        output_raster.SetGeoTransform(geotransform)
        out_band = output_raster.GetRasterBand(1)
        out_band.WriteArray(self.grid)
        out_band.FlushCache()
        out_band.SetNoDataValue(-99)
        del output_raster

    def plot(self):  # pragma: no cover
        """
        Returns a matplotlib.pyplot.figure object containing an image of the fragmented landscape (with axes removed).

        Requires that the fragmented landscape has been created already using :meth:`~create`.

        :return: figure object containing the fragmented landscape.
        :rtype: matplotlib.pyplot.figure
        """
        fig = plt.figure()
        fig.figimage(self.grid, cmap="Greys_r", resize=True)
        return fig
